/*********************************************************************
*                   (c) SEGGER Microcontroller GmbH                  *
*                        The Embedded Experts                        *
*                           www.segger.com                           *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose : Implementation of low-level functions for I/O with the
          SEGGER Runtime Library
          using a UART (SEGGER's BSP UART module)
*/

/*********************************************************************
*
*       #include section
*
**********************************************************************
*/

#include "__SEGGER_RTL_Int.h"
#include "stdio.h"
#include "BSP_UART.h"

/*********************************************************************
*
*       Defines, fixed
*
**********************************************************************
*/

#define _PRINOPS_UART_RINGBUFFER_SIZE 64

/*********************************************************************
*
*       Prototypes
*
**********************************************************************
*/
void RTL_UART_Init(unsigned int Unit, unsigned long Baudrate, unsigned char NumDataBits, unsigned char Parity, unsigned char NumStopBits);

/*********************************************************************
*
*       Local types
*
**********************************************************************
*/

struct __SEGGER_RTL_FILE_impl {         // NOTE: Provides implementation for FILE
  int stub; // only needed so impl has size != 0.
};

//
// Simple ringbuffer implementation. Buffer is empty if Head == Tail.
//
struct _RINGBUFFER {
  char Buf[_PRINOPS_UART_RINGBUFFER_SIZE];
  unsigned Head;  // current writer position
  unsigned Tail;  // current reader position
};

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

static FILE __SEGGER_RTL_stdin_file  = { 0 };  // stdin reads from UART
static FILE __SEGGER_RTL_stdout_file = { 0 };  // stdout writes to UART
static FILE __SEGGER_RTL_stderr_file = { 0 };  // stderr writes to UART

static unsigned int                _UART_Port   = 0;
static volatile int                _UART_Busy   = 0;
static int                         _stdin_ungot  = EOF;
static volatile struct _RINGBUFFER _UART_OutBuf = { 0 };
static volatile struct _RINGBUFFER _UART_InBuf  = { 0 };

/*********************************************************************
*
*       Public data
*
**********************************************************************
*/

FILE *stdin  = &__SEGGER_RTL_stdin_file;  // NOTE: Provide implementation of stdin for RTL.
FILE *stdout = &__SEGGER_RTL_stdout_file; // NOTE: Provide implementation of stdout for RTL.
FILE *stderr = &__SEGGER_RTL_stderr_file; // NOTE: Provide implementation of stderr for RTL.

/*********************************************************************
*
*       Static code
*
**********************************************************************
*/

/*********************************************************************
*
*       _Ringbuffer_Empty()
*
*  Function description
*    Checks if given ringbuffer is empty.
*
*  Parameters
*    Buf:   pointer to ringbuffer
*
*  Return value
*    != 0: Buffer empty
*    == 0: There is some data in the buffer
*/
static inline int _Ringbuffer_Empty(volatile struct _RINGBUFFER *Buf) {
  return Buf->Head == Buf->Tail;
}

/*********************************************************************
*
*       _Ringbuffer_Full()
*
*  Function description
*    Checks if given ringbuffer is full.
*
*  Parameters
*    Buf:   pointer to ringbuffer
*
*  Return value
*    != 0: Buffer full
*    == 0: There is some space in the buffer
*/
static inline int _Ringbuffer_Full(volatile struct _RINGBUFFER *Buf) {
  return Buf->Tail == ((Buf->Head+1) % _PRINOPS_UART_RINGBUFFER_SIZE);
}

/*********************************************************************
*
*       _Ringbuffer_Put1IfFree()
*
*  Function description
*    If there is some free space, put a single char into the ringbuffer.
*    Always returns immediately.
*
*  Parameters
*    Buf:   pointer to ringbuffer
*    p:     char to put into buffer
*
*  Return value
*    == 0: Buffer was full, nothing was stored.
*    == 1: Character was stored in buffer.
*/
static size_t _Ringbuffer_Put1IfFree(volatile struct _RINGBUFFER *Buf, char p) {
  if (_Ringbuffer_Full(Buf)) {
    return 0;
  } else {
    Buf->Buf[Buf->Head] = p;
    Buf->Head = (Buf->Head+1) % _PRINOPS_UART_RINGBUFFER_SIZE;
    return 1;
  }
}

/*********************************************************************
*
*       _Ringbuffer_Get1()
*
*  Function description
*    Blockingly pull a single char out of a ringbuffer.
*
*  Parameters
*    Buf:   pointer to ringbuffer
*
*  Return Value
*    Character that was taken from buffer
*/
static char _Ringbuffer_Get1(volatile struct _RINGBUFFER *Buf) {
  char c;

  while (_Ringbuffer_Empty(Buf)) {
    //
    // busy wait until there is data in ringbuffer
    //
  }
  c = Buf->Buf[Buf->Tail];
  Buf->Tail = (Buf->Tail+1) % _PRINOPS_UART_RINGBUFFER_SIZE;
  return c;
}

/*********************************************************************
*
*       _UART_cbOnRxComplete()
*
*  Function description
*    RX-complete callback for UART.
*    Called when a character has been received.
*
*  Parameters
*    Unit:    UART Port
*    c:       received character
*/
static void _UART_cbOnRxComplete(unsigned int Unit, unsigned char c) {
  if (Unit == _UART_Port) {
    _Ringbuffer_Put1IfFree(&_UART_InBuf, c);
  }
}

/*********************************************************************
*
*       _UART_Transmit1
*
*  Function description
*    Immediately transmit single char from ringbuffer via UART.
*
*  Additional information
*    Must only be called if UART is not busy, yet.
*    Will block if TX buffer is empty, so should only be called if
*    buffer is NOT empty.
*/
static void _UART_Transmit1(void) {
  char c;

  _UART_Busy = 1;
  c = _Ringbuffer_Get1(&_UART_OutBuf);
  BSP_UART_Write1(_UART_Port, c);
}

/*********************************************************************
*
*       _UART_cbOnTxComplete()
*
*  Function description
*    TX-complete callback for UART.
*    Called when sending a character has finished.
*
*  Parameters
*    Unit       : UART Port
*
*  Return value
*    == 0: There are more bytes to be sent
*    != 0: Buffer empty, no more bytes to be sent.
*/
static int _UART_cbOnTxComplete(unsigned int Unit) {
  if (Unit == _UART_Port) {
    _UART_Busy = 0;
    if (!_Ringbuffer_Empty(&_UART_OutBuf)) {
      _UART_Transmit1();
      return 0;
    }
  }
  return 1;
}

/*********************************************************************
*
*       _stdin_getc()
*
*  Function description
*    Get character from standard input.
*
*  Return value
*    Character received.
*
*  Additional information
*    This function never fails to deliver a character.
*/
static char _stdin_getc(void) {
  char c;

  if (_stdin_ungot != EOF) {
    c = _stdin_ungot;
    _stdin_ungot = EOF;
  } else {
    c = _Ringbuffer_Get1(&_UART_InBuf);
  }
  return c;
}

/*********************************************************************
*
*       Public code
*
**********************************************************************
*/

/*********************************************************************
*
*       RTL_UART_Init()
*
*  Function description
*    Initialize RTL to use given UART for stdio.
*
*  Parameters
*    Unit       : UART unit number (typically zero-based).
*    Baudrate   : Baud rate to configure [Hz].
*    NumDataBits: Number of data bits to use.
*    Parity     : One of the following values:
*                   * BSP_UART_PARITY_NONE
*                   * BSP_UART_PARITY_ODD
*                   * BSP_UART_PARITY_EVEN
*    NumStopBits: Number of stop bits to use.
*
*  Additional description
*    Parameters are same as for BSP_UART_Init().
*    This also sets appropriate RX and TX interrupt handlers.
*/
void RTL_UART_Init(unsigned int Unit, unsigned long Baudrate, unsigned char NumDataBits, unsigned char Parity, unsigned char NumStopBits) {
  _UART_Port = Unit;
  BSP_UART_Init(_UART_Port, Baudrate, NumDataBits, Parity, NumStopBits);
  BSP_UART_SetReadCallback(_UART_Port, _UART_cbOnRxComplete);
  BSP_UART_SetWriteCallback(_UART_Port, _UART_cbOnTxComplete);
}

/*********************************************************************
*
*       __SEGGER_RTL_X_file_stat()
*
*  Function description
*    Get file status.
*
*  Parameters
*    stream - Pointer to file.
*
*  Additional information
*    Low-overhead test to determine if stream is valid.  If stream
*    is a valid pointer and the stream is open, this function must
*    succeed.  If stream is a valid pointer and the stream is closed,
*    this function must fail.
*
*    The implementation may optionally determine whether stream is
*    a valid pointer: this may not always be possible and is not
*    required, but may assist debugging when clients provide wild
*    pointers.
*
*  Return value
*    <  0 - Failure, stream is not a valid file.
*    >= 0 - Success, stream is a valid file.
*/
int __SEGGER_RTL_X_file_stat(FILE *stream) {
  if (stream == stdin || stream == stdout || stream == stderr) {
    return 0;       // NOTE: stdin, stdout, and stderr are assumed to be valid.
  } else {
    return EOF;
  }
}

/*********************************************************************
*
*       __SEGGER_RTL_X_file_bufsize()
*
*  Function description
*    Get stream buffer size.
*
*  Parameters
*    stream - Pointer to file.
*
*  Additional information
*    Returns the number of characters to use for buffered I/O on
*    the file stream.  The I/O buffer is allocated on the stack
*    for the duration of the I/O call, therefore this value should
*    not be set arbitrarily large.
*
*    For unbuffered I/O, return 1.
*
*  Return value
*    Nonzero number of characters to use for buffered I/O; for
*    unbuffered I/O, return 1.
*/
int __SEGGER_RTL_X_file_bufsize(FILE *stream) {
  (void)stream;
  return 1;  // our ringbuffer takes care of that.
}

/*********************************************************************
*
*       __SEGGER_RTL_X_file_read()
*
*  Function description
*    Read data from file.
*
*  Parameters
*    stream - Pointer to file to read from.
*    s      - Pointer to object that receives the input.
*    len    - Number of characters to read from file.
*
*  Return value
*    >= 0 - Success, amount of data read.
*    <  0 - Failure.
*
*  Additional information
*    Reading from any stream other than stdin results in an error.
*/
int __SEGGER_RTL_X_file_read(FILE *stream, char *s, unsigned len) {
  int c;

  if (stream == stdin) {
    c = 0;
    while (len > 0) {
      *s = _stdin_getc();
      ++s;
      ++c;
      --len;
    }
  } else {
    c = EOF;
  }
  return c;
}

/*********************************************************************
*
*       __SEGGER_RTL_X_file_write()
*
*  Function description
*    Write data to file.
*
*  Parameters
*    stream - Pointer to file to write to.
*    s      - Pointer to object to write to file.
*    len    - Number of characters to write to the file.
*
*  Return value
*    >= 0 - Success.
*    <  0 - Failure.
*
*  Additional information
*    this version is NOT reentrant!
*    a reentrant version needs to lock use of the output ring buffer and _UART_Busy.
*    stdout and stderr are directed to UART;
*    writing to any stream other than stdout or stderr results in an error
*/
int __SEGGER_RTL_X_file_write(FILE *stream, const char *s, unsigned len) {
  unsigned p;
  int r;

  if ((stream == stdout) || (stream == stderr)) {
    for (p = 0; p < len; ++p) {
      do {
        r = _Ringbuffer_Put1IfFree(&_UART_OutBuf, s[p]);
      } while (r < 1);
      if (!_UART_Busy) {
        _UART_Transmit1();
      }
    }
    return len;
  } else {
    return EOF;
  }
}

/*********************************************************************
*
*       __SEGGER_RTL_X_file_unget()
*
*  Function description
*    Push character back to stream.
*
*  Parameters
*    stream - Pointer to file to push back to.
*    c      - Character to push back.
*
*  Return value
*    >= 0 - Success.
*    <  0 - Failure.
*
*  Additional information
*    Push-back is only supported for standard input, and
*    only a single-character pushback buffer is implemented.
*/
int __SEGGER_RTL_X_file_unget(FILE *stream, int c) {
  if (stream == stdin) {
    if (c != EOF && _stdin_ungot == EOF) {
      _stdin_ungot = c;
    } else {
      c = EOF;
    }
  } else {
    c = EOF;
  }
  return c;
}

/*************************** End of file ****************************/
